"""Various utils related to handling file system paths"""
import itertools
import pathlib


def find_upwards(start: pathlib.Path, name: str) -> pathlib.Path:
    for parent in itertools.chain([start], start.parents):
        path = parent / name
        if path.exists():
            return path
    raise FileNotFoundError



"""Lets the user specify path dependend aliases"""
from __future__ import annotations

import ast
import dataclasses
import pathlib
import shlex
import subprocess
from typing import Dict, Iterable, List, Optional, Tuple

from kindly import _pathutils


@dataclasses.dataclass(frozen=True)
class AliasCommand:
    cwd: pathlib.Path
    name: str
    help: Optional[str]
    steps: Tuple[Tuple[str, ...], ...]

    def __call__(self, args: List[str]) -> None:
        for step in self.steps:
            subprocess.check_call(step + tuple(args), cwd=self.cwd)

    @staticmethod
    def from_spec(cwd: pathlib.Path, name: str, spec: Dict) -> AliasCommand:
        try:
            raw_steps = spec["steps"]
        except KeyError:
            raw_steps = [spec["step"]]

        steps = [
            shlex.split(step) if isinstance(step, str) else step for step in raw_steps
        ]

        # Would be nice to extend this to allow containerized commands akin to jobs in
        # github workflows.

        return AliasCommand(
            cwd=cwd,
            name=name,
            help=spec.get("help", name.capitalize().replace("_", " ")),
            steps=tuple(map(tuple, steps)),
        )


class AliasesProvider:
    # pylint: disable=too-few-public-methods

    def __init__(self, cwd: pathlib.Path) -> None:
        self._cwd = cwd

    def v1_commands(self) -> Iterable[AliasCommand]:
        try:
            kindly_pyi = _pathutils.find_upwards(
                pathlib.Path.cwd(), "kindly_aliases.pyl"
            )
        except FileNotFoundError:
            return

        raw_commands = ast.literal_eval(kindly_pyi.read_text())["commands"]
        for name, spec in raw_commands.items():
            yield AliasCommand.from_spec(self._cwd, name, spec)



import argparse
import dataclasses
import logging
import os
import pathlib
import subprocess
from textwrap import dedent
from typing import Any, List, Optional

import argcomplete
import pkg_resources

from kindly import api

logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class _V1Adapter:
    v1_command: api.V1Command

    @property
    def name(self) -> str:
        return self.v1_command.name

    @property
    def help(self) -> Optional[str]:
        return self.v1_command.help

    def configure_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("args", nargs="*")

    def __call__(self, args: argparse.Namespace) -> None:
        self.v1_command(args.args)


def _parser(cwd: pathlib.Path) -> argparse.ArgumentParser:
    program_parser = argparse.ArgumentParser(
        description=dedent(
            """\
            Enable tab completion by running
            `eval "$(register-python-argcomplete kindly)"`.
            """
        )
    )
    # TODO: Fix crash when no arguments are provided and no commands exist
    program_subparsers = program_parser.add_subparsers(required=True)

    # TODO: Define strategy for dealing with conflicts
    # Maybe enable the user to disable (distribution,command) tuples in the tools
    # section of pyproject.toml.
    # One distribution could still provide the same command multiple times (this
    # distribution provides any command that the user specifies in the aliases
    # file); this should probably be an error.

    for entry_point in pkg_resources.iter_entry_points("kindly.provider"):
        cls = entry_point.load()
        provider: api.Provider = cls(cwd)

        # TODO: Better type hints
        commands: dict[str, Any] = {}

        if hasattr(provider, "v1_commands"):
            for command in provider.v1_commands():
                commands[command.name] = _V1Adapter(command)

        if hasattr(provider, "v2_commands"):
            for command in provider.v2_commands():  # type: ignore
                commands[command.name] = command

        for name, command in sorted(commands.items()):
            command_parser = program_subparsers.add_parser(
                name,
                help=name.capitalize().replace("_", " ")
                if command.help is None
                else command.help,
            )
            command.configure_parser(command_parser)  # type: ignore
            command_parser.set_defaults(func=command)

    return program_parser


def cli(cwd: pathlib.Path, args: List[str]) -> None:
    # noinspection PyUnresolvedReferences
    # pylint: disable=protected-access
    logging.basicConfig(level=logging._nameToLevel[os.environ.get("LEVEL", "WARNING")])
    parser = _parser(cwd)
    argcomplete.autocomplete(parser)
    parsed = parser.parse_args(args)
    try:
        parsed.func(parsed)
    except subprocess.SubprocessError:
        parser.exit(1)



"""Specification of public interface for plugins"""
from __future__ import annotations

import pathlib
import sys
from typing import Iterable, List, Optional

if sys.version_info[1] < 8:
    from typing_extensions import Protocol
else:
    from typing import Protocol


class V1Command(Protocol):
    @property
    def name(self) -> str:
        """Return name of command"""

    @property
    def help(self) -> Optional[str]:
        """Return help message, if any

        If None is returned the help message will be chosen by kindly.
        """

    def __call__(self, args: List[str]) -> None:
        """Implementation of the command

        Receives any arguments that were not consumed by parsing.
        """


class Provider(Protocol):
    # pylint: disable=too-few-public-methods
    def __init__(self, cwd: pathlib.Path) -> None:
        ...

    def v1_commands(self) -> Iterable[V1Command]:
        ...
